
/*	
	OligoTools.c
	
	This is a collection of tools to be used with primer design / analysis
	Developped by James Fraser 28/03/08
	
*/

#include <strings.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>



/* Used to calculate the primer with the optimum Tm */
/* Tables used were from Breslauer et Al., "Predicting DNA duplex stability from the base sequence" Proc. Natl. Acad. Sci. USA, June 1986, Vol 83, pp. 3746-3750 */
/* Formula used to calculate Tm is from W. Rychlik et al., "Optimization of the annealing temperature for DNA amplification in vitro", Nucleic Acids Research, August 1990, Vol 18 .No 21, pp 6409-6412 */
double Tm(char *oligo)
{
	
	int start, size;
	double h = 0.0, s = 10.8, g = -5.0, Temp, c, K;
	/* Delta G does not start at zero because a symmetry term must be included in the calculation.
	*  This value of -5.0 stems from the helix initiation free energy (5) and Delta Gsym 
	* Delta S has an initiation value */
	char current, next;
		
	start = strlen(oligo) -1;
	
	/* This part may be confusing: current is the 5' base, next is the 3' base. We're working backwards from the 3' end */
	current = oligo[start-1];
	next = oligo[start];
	
	for(size = 0; size < start; size++) {
		
		if ( (current == 'A' && next == 'A') || (current == 'T' && next == 'T') ) {
		
			h = h + 9.1;
			s = s + 24.0;
			g = g + 1.9;
		
		} else if (current == 'A' && next == 'T') {
		
			h = h + 8.6;
			s = s + 23.9;
			g = g + 1.5;
			
		} else if ( (current == 'A' && next == 'C') || (current == 'G' && next == 'T')) {
		
			h = h + 6.5;
			s = s + 17.3;
			g = g + 1.3;
			
		} else if ( (current == 'A' && next == 'G') || (current == 'C' && next == 'T') ) {
		
			h = h + 7.8;
			s = s + 20.8;
			g = g + 1.6;
		 
		} else if ( (current == 'C' && next == 'A') || (current == 'T' && next == 'G')) {
		
			h = h + 5.8;
			s = s + 12.9;
			g = g + 1.9;

		} else if ( (current == 'C' && next == 'C') || (current == 'G' && next == 'G') ) {
		
			h = h + 11.0;
			s = s + 26.6;
			g = g + 3.1;

		} else if (current == 'C' && next == 'G') {
		
			h = h + 11.9;
			s = s + 27.8;
			g = g + 3.6;
				
		} else if ( (current == 'G' && next == 'A') || (current == 'T' && next == 'C') ) {
		
			h = h + 5.6;
			s = s + 13.5;
			g = g + 1.6;
		
		} else if (current == 'G' && next == 'C') {
		
			h = h + 11.1;
			s = s + 26.7;
			g = g + 3.1;

		} else if (current == 'T' && next == 'A') {
		
			h = h + 6.0;
			s = s + 16.9;
			g = g + 0.9;
							
		} else {
				
			printf("%c%c  Delta G: %f\n", current, next, g);
		
		}
		
		next = current;
		current = oligo[start-2-size];
			
	}	

	h = h * -1000.0;
	s = -s;
	c = 200.0 / 1000000000;
	K = 50.0 / 1000;
	
	Temp = ( ( h ) / ( s + (1.987 * log( c / 4.0 ) ) ) ) - 273.15 + 16.6 * log10( K );
	
	return Temp;
	
}

/* Returns the annealing temperature of the entered oligo using the simple calculation */
int TmSimple(char *oligo) {

	int i, Tm = 0;

	/* Loops through the string */
	for(i = 0; oligo[i] != '\0'; i++) {
	
		/*Checks what base is present and adjusts Tm */
		if (oligo[i] == 'A' || oligo[i] == 'T') {
		
			Tm = Tm + 2;
		
		} else {
		
			Tm = Tm + 4 ;
		
		}
			
	}

	return Tm;

}

/* returns the maximum number of base pairs made if the primer were to form a hairpin */
int hairpin(char *oligo) {
	
	int length = strlen(oligo) -1; /* The -1 is there so that length can be used in the array */
	int i, mid, op, x = 0, max = 0;

	/* goes through the entire primer, using mid as the location where the hairpin forms */
	for(mid = 2; mid < (length - 1); mid ++) {
			
		/* Used for hairpins with three base turns; mid is the location of the middle of the hairpin structure */
		for(i = 0; i < (mid - 1); i++ ) {
		
			/* op is the location of the opposite base to i */
			op = mid + mid - i;
			
			/* Occurs if base has nothing to pair to */
			if ( op > length) {
			
				/* Do nothing */
			
			} else if ( (oligo[i] == 'A' && oligo[op] == 'T') || (oligo[i] == 'T' && oligo[op] == 'A') || (oligo[i] == 'C' && oligo[op] == 'G') || (oligo[i] == 'G' && oligo[op] == 'C') ) {
							
				x++;
				
			}
					
		}
		
		/* Changes max if the number of base pairs at this mid (Or hairpin conformation) is higher than the previous max */
		if ( x > max) {
		
			max = x;
			
		}
		
		/* Resets x */
		x = 0;
		
		/* Used for hairpins with four base turns with the middle between mid and mid + 1 */
		for(i = 0; i < (mid - 1); i++ ) {
		
			/* op is the location of the opposite base to i */
			op = mid + mid - i + 1;
			
			/* Occurs if base has nothing to pair to */
			if ( op > length) {
			
				/* Do nothing */
			
			} else if ( (oligo[i] == 'A' && oligo[op] == 'T') || (oligo[i] == 'T' && oligo[op] == 'A') || (oligo[i] == 'C' && oligo[op] == 'G') || (oligo[i] == 'G' && oligo[op] == 'C') ) {
							
				x++;
				
			}
					
		}
		
		/* Changes max if the number of base pairs at this mid (Or hairpin conformation) is higher than the previous max */
		if ( x > max) {
		
			max = x;
			
		}
		
		/* Resets x */
		x = 0;
		
	}

	return max;

}

/* Returns the number of cycles it takes to make the oligo */
/* Expects a 5' - 3' oligo, and synthesis occurs from 3' to 5' */
int oligo_cycles(char *oligo) {

	int cycles = 0, i = 0, length = strlen(oligo);
	char order[4];
	order[0] = 'A', order[1] = 'C', order[2] = 'G', order[3] = 'T';
	
	while ( i < length ) {
	
		if ( oligo[ length - i -1] == order[cycles%4] ) 
		{ 
			i++;
			
		} else if (oligo[ length - i -1] == 'N') {
		
			i++;
		
		}
		
		cycles++;
	
	}
	
	return cycles;
	
}

// Returns the % GC content for an oligo
double GC_Content(char *oligo) {

	int size = strlen(oligo), i;
	int count = 0;
	double GC;
	
	for (i = 0; i <= size; i++) {
	
		if (oligo[i] == 'G' || oligo[i] == 'C') {
				
			count++;
			
		}
			
	}
	
	GC = (double)count / (size);

	
	
	return GC;

}

//Used for sorting
int comp_nums(const double *num1, const double *num2)
{
  if (*num1 <  *num2) {
	return 1;	
	} else if (*num1 == *num2){
		  return  0;
	} else if (*num1 >  *num2) {
	
	return  -1;
	
	}
	
	return 100;
}

//Opens a file and extracts DNA. Returns 0 the number of bases imported
//Form: importDNA(Filename to be searched, Array for DNA, String for header)
int importDNA(char *filename, char *DNA, char *head)
{
	int i =0,n ;
	char y, z;
	FILE *file;
		
	file = fopen(filename, "r");
	if (file == NULL) {
			
		printf("\nError: Can't open the file \"%s\"!.\n", filename);
		return 0;
			
	}

	printf("\nOpening code...%s\n", filename);

	/* Extracts Header from file. Header is the first line of code */
	if ( (z = (char)fgetc(file)) == '>') {
	
		while ( z != '\n' && z != '\r')	
		{
		
			head[i] = z;
			i++;
			z = (char)fgetc(file);
		
		}
		
		head[i] = '\0'; /* Makes header a string */
		
	} else {
		
		strcpy(head, "(none)");
		rewind(file);
		
	}
		
	/* enters the code into memory, only takes ACTG values */
	for (n = 0; feof(file) == 0; n++) {
			
		y = (char)fgetc(file); 
			
		/* Skips anything other than ACTG/acgt, which is entered into an array. Complementary strand also generated */
		if (y == 'A' || y == 'a') {
		
			DNA[n] = 'A';
	
		} else if (y == 'C' || y == 'c') {
			
			DNA[n] = 'C';
	
		} else if (y == 'G' || y == 'g') {
				
			DNA[n] = 'G';
			
		} else if (y == 'T' || y == 't') { 
				
			DNA[n] = 'T';
			
		} else if (y == 'N') {
		
			DNA[n] = 'N';
		
		} else {
					
			/*Does nothing; skips to the next char */
			n--; /*Used to correct n such that it represents the number of bases in the code */
					
		}
		
	}
		
	printf("%d bases of code has been entered into memory.\n", n);
	fclose(file);
	

	return n;

}


//Takes strand of DNA and reverse complements it.
// complement(DNA to be reversed, array for complementary DNA)
void complement(char *DNA, char *complement)
{
	int i;
	int length = strlen(DNA)-1;
	
	for (i = 0; i<=length; i++) {
		
		if (DNA[length-i] == 'A') {
		
			complement[i] = 'T';
	
		} else if (DNA[length-i] == 'C') {
			
			complement[i] = 'G';
	
		} else if (DNA[length-i] == 'G') {
				
			complement[i] = 'C';
			
		} else if (DNA[length-i] == 'T') { 
				
			complement[i] = 'A';
			
		} else if (DNA[length-i] == 'N') {
		
			complement[i] = 'N';
		
		} else {
					
			complement[i] = 'X'; //Represents that an unknown character was present
				
		}
		  
	} 
	
	complement[i] = '\0';
	
}

//Cuts DNA based on enzyme
// cutter(Array of DNA to be used, Enzyme sequence, number of bases, Array of cuts)
void cutter(char *DNA, char *enzyme, int n, int *cut ) {
	
	int i, a=0, z, count;
	
	printf("Scanning for restriction sites...\n");
	for (i = 0; i < (n - 6) ; i++) {
	
		count = 0;
					
		/* Marks the location of any cut sites in the array cut */
		for (z = 0; z< strlen(enzyme); z++) {
		
			if (DNA[i+z] == enzyme[z]) {count++;}

		}
		
		//printf("COUNT: %s\n", count);
		
		if (count == strlen(enzyme)){ // Complete match
			
			cut[a] = i + (strlen(enzyme)/2) - 1;
			a++;
		
		}
	
	}
	
	/* Checks that there are cut sites in the sequence, and exits if there are none */
	if ( a == 0) {
	
		printf("No cut sites were found!\n");
	
	} else {
	
		printf("Completed scan: %d cut sites found!\n", a);
	
	}
	
	/* Puts null character in to mark end of array */
	cut[a] = '\0';

}

//Runs the RepeatMasker program
int RepeatMasker(char *input, char *masked_dna, char *organism, char *program_path) {

	char Repeat[1000];
	char head[200];
		
	//Check to make sure program actually exists / path is correct
	if (file_exists(program_path)== EXIT_FAILURE) {
		
		printf("The path does not exist! Doublecheck that the path points to the right location:\n%s\n", program_path);
		return EXIT_FAILURE;
		
	}
	

	sprintf(Repeat, "%s -pa 2 -q -species %s %s", program_path, organism, input);
	system(Repeat);
	
	sprintf(Repeat, "%s.masked", input);
	importDNA(Repeat,masked_dna,head);
	
	//Remove files
	sprintf(Repeat, "%s.cat", input);
	unlink(Repeat);
	sprintf(Repeat, "%s.out", input);
	unlink(Repeat);
	sprintf(Repeat, "%s.tbl", input);
	unlink(Repeat);
	sprintf(Repeat, "%s.masked", input);
	unlink(Repeat);	
	
	return EXIT_SUCCESS;
	

}

//Froma  given segment of DNA, store the characters between the two indices in output.
void extract(char *DNA, int index1, int index2, char *output) {
	
	int i;
		
	for (i = index1; i<= index2; i++) {
	
		output[i-index1] = DNA[i];
			
	}
	
	output[i-index1] = '\0';

}

//returns number of characters
int counter(char *input, char c) {

	int i,j = 0;
	
	for (i = 0; i < strlen(input); i++) {
	
		if (input[i] == c){
		
			j++;
			
		}
	
	}
	
	return j;

}

//Checks if a file exists or not.
int file_exists(char *filename)
{
	FILE *file;
    if ( file = fopen(filename, "r"))
    {
        fclose(file);
        return EXIT_SUCCESS;
    }
    return EXIT_FAILURE;
}




